use crate::schema;
use tide::http::Cookie;

#[derive(Clone)]
pub struct ServerState {
    core: &'static super::ServerCoreState,
}

type Request = tide::Request<ServerState>;

async fn v1_check(req: Request) -> tide::Result<tide::Response> {
    let validity = super::is_auth_valid(req.state().core, &req);
    if let Some(true) = validity {
        return Ok(tide::Response::builder(200).build());
    } else if let Some(false) = validity {
        return Ok(tide::Response::builder(403).build());
    } else {
        return Ok(tide::Response::builder(401).build());
    }
}

fn build_session(
    qi: &microrm::QueryInterface,
    response: &mut tide::Response,
) -> tide::Result<schema::SessionID> {
    let rng = ring::rand::SystemRandom::new();
    let session_id: [u8; 32] = ring::rand::generate(&rng)
        .map_err(|_| tide::Error::from_str(500, "Failed to generate session ID"))?
        .expose();
    let session_id = base64::encode_config(session_id, base64::URL_SAFE_NO_PAD);

    response.insert_cookie(Cookie::new("vogt_session", session_id.clone()));

    let maybe_id = qi.add(&crate::schema::Session { key: session_id });
    Ok(maybe_id.ok_or(tide::Error::from_str(
        500,
        "Failed to store session in database",
    ))?)
}

fn render_login_page(state: &ServerState, mut response: tide::Response, to_present: schema::AuthChallengeType) -> tide::Response {
    let tmpl = state.core.templates.read().unwrap();

    let do_challenge = |ty,ch| {
        tmpl.render("id_v1_login", &serde_json::json!(
            {
                "challenge":
                    format!(r#"
                        <input type="hidden" name="challenge-type" value="{:?} {}" />
                        <div class="challenge-type">{}</div>
                        <div class="challenge-content">{}</div>
                        "#,
                        to_present, serde_json::to_string(&to_present).unwrap(),ty,ch)
            }
        )).unwrap()
    };

    response.set_content_type("text/html");

    match to_present {
        schema::AuthChallengeType::Username => {
            response.set_body(do_challenge("Username",
                r#"<input name="challenge" type="text" autofocus />"#));
        },
        schema::AuthChallengeType::Password => {
            response.set_body(do_challenge("Password",
                r#"<input name="challenge" type="password" autofocus />"#));
        }
        _ => todo!()
    }

    response
}

async fn v1_login(mut req: tide::Request<ServerState>) -> tide::Result<tide::Response> {
    let mut response = tide::Response::builder(200).build();
    let qi = req.state().core.pool.query_interface();

    let sid: schema::SessionID = match req.cookie("vogt_session") {
        Some(sid) => {
            let existing = qi.get_one_by(schema::SessionColumns::Key, sid.value());

            if let Some(id) = existing {
                id.id()
            } else {
                build_session(&qi, &mut response)?
            }
        }
        None => build_session(&qi, &mut response)?,
    };

    let realm_name = req
        .param("realm")
        .expect("Failed to parse realm out of path?");

    let realm_id = qi
        .get_one_by(schema::RealmColumns::Shortname, realm_name)
        .ok_or(tide::Error::from_str(404, "No such realm"))?;

    let realm_auth = qi.get_one_by_multi(
        &[
            schema::SessionAuthenticationColumns::Session,
            schema::SessionAuthenticationColumns::Realm,
        ],
        &microrm::value_list!(&sid, &realm_id.id()),
    );

    let to_present: Option<schema::AuthChallengeType> = match realm_auth {
        None => Some(schema::AuthChallengeType::Username),
        Some(auth) => auth.challenges_left.first().map(|x| *x),
    };

    if to_present.is_some() {
        Ok(render_login_page(req.state(), response, to_present.unwrap()))
    }
    else {
        // already logged in...
        todo!()
    }
}

async fn v1_login_response(mut req: tide::Request<ServerState>) -> tide::Result<tide::Response> {
    todo!()
}

pub fn id_v1_server(core: &'static super::ServerCoreState) -> tide::Server<ServerState> {
    let mut srv = tide::with_state(ServerState { core });

    srv.with(tide::log::LogMiddleware::new());

    srv.at("login").get(v1_login).post(v1_login_response);
    srv.at("check").get(v1_check).head(v1_check);

    // load and register templates
    core.templates.write().unwrap()
        .register_template_file("id_v1_login", "srv/login.html")
        .expect("Couldn't register template!");

    srv
}
